@@ -1,5 +1,7 @@
 use crate::errors;
+use log::*;
 use rand::seq::IteratorRandom;
+use rayon::prelude::*;
 use realfft::RealFftPlanner;
 use rustfft::num_complex::Complex;
 use serde::Serialize;
@@ -16,8 +18,7 @@ pub fn ad_dual(sample: &[f64], other: &[f64]) -> f64 {
     let k = lnx + lny;
     let mut ad_i = Vec::new();
     for i in x {
-        let step: Vec<f64> = sample.iter().filter(|z| **z <= i).map(|z| *z).collect();
-        ad_i.push(step.len() as f64);
+        ad_i.push(sample.iter().filter(|z| **z <= i).count() as f64);
     }
     let mut adi = Vec::new();
 
@@ -51,9 +52,8 @@ pub fn cdf(x: &[f64]) -> Vec<(f64, f64)> {
     cx.dedup();
     let mut cdf = Vec::new();
     for i in cx {
-        let num: Vec<&f64> = x.iter().filter(|x| **x <= i).collect();
-        let res = num.len() as f64 / ln;
-        cdf.push((i, res));
+        let num = x.iter().filter(|x| **x <= i).count() as f64 / ln;
+        cdf.push((i, num));
     }
     cdf
 }
@@ -96,11 +96,9 @@ pub fn cdf_dual(obs: &[f64], other: &[f64]) -> Vec<(f64, f64)> {
     // construct cdf of each
     let mut cdf = Vec::new();
     for i in x {
-        let numx: Vec<&f64> = xo.iter().filter(|z| **z <= i).collect();
-        let numy: Vec<&f64> = yo.iter().filter(|z| **z <= i).collect();
-        let resx = numx.len() as f64 / lnx;
-        let resy = numy.len() as f64 / lny;
-        cdf.push((resx, resy));
+        let numx = xo.iter().filter(|z| **z <= i).count() as f64 / lnx;
+        let numy = yo.iter().filter(|z| **z <= i).count() as f64 / lny;
+        cdf.push((numx, numy));
     }
     cdf
 }
@@ -212,6 +210,28 @@ pub fn cdf_rng(obs: &[f64], range: &std::ops::Range<i32>) -> Vec<f64> {
     cdf_rng
 }
 
+/// Chi-squared goodness-of-fit test.
+pub fn chi_squared(obs: &[f64], other: &[f64]) -> f64 {
+    // join the two vectors, sort and deduplicate
+    let mut x = obs.to_vec();
+    let mut y = other.to_vec();
+    let xo = x.clone(); // clone the originals for later use
+    let yo = y.clone();
+    x.append(&mut y);
+    x.sort_by(|a, b| a.partial_cmp(b).unwrap());
+    x.dedup();
+    // construct cdf of each
+    let mut difs = Vec::new();
+    for i in x {
+        let numx = xo.iter().filter(|z| **z <= i).count() as f64;
+        let numy = yo.iter().filter(|z| **z <= i).count() as f64;
+        if numy > 0.0 {
+            difs.push(f64::powi(numx - numy, 2) / numy);
+        }
+    }
+    difs.iter().sum::<f64>()
+}
+
 /// Convolve frequency distributions along an index of values.
 ///
 /// # Examples
@@ -296,6 +316,76 @@ pub fn convo(x: &[f64], y: &[f64], rng: i32) -> Vec<f64> {
     out_data
 }
 
+/// Less convoluted convolution.
+pub fn convo_vec(obs: Vec<Vec<f64>>, index: std::ops::Range<i32>) -> Vec<f64> {
+    let mut out = vec![Complex::new(0.0, 0.0); index.len()];
+    let mut real_planner = RealFftPlanner::<f64>::new();
+    info!("Constructing FftPlanner for fourier transforms.");
+    let r2c = real_planner.plan_fft_forward(index.len() + 1);
+    for ob in obs {
+        let cdf = cdf_rng(&ob, &index);
+        let mut pmf = pmf_from_cdf(&cdf);
+        let mut spectrum = r2c.make_output_vec();
+        info!("Forward transforming the pmf using fft.");
+        r2c.process(&mut pmf, &mut spectrum).unwrap();
+        info!("Summing transformed pmfs.");
+        out = out
+            .iter()
+            .zip(spectrum.iter())
+            .map(|(a, b)| a + b)
+            .collect::<Vec<Complex<f64>>>();
+    }
+    info!("Constructing FftPlanner for inverse fourier transforms.");
+    let c2r = real_planner.plan_fft_inverse(index.len() + 1);
+    info!("Inverse fourier transform using fft.");
+    let mut out_data = c2r.make_output_vec();
+    c2r.process(&mut out, &mut out_data).unwrap();
+    info!("Normalize output by dividing by length.");
+    out_data = out_data
+        .par_iter()
+        .map(|a| a / index.len() as f64)
+        .collect::<Vec<f64>>();
+    info!("Normalize output by dividing by sum of pmfs (having added several pmfs together).");
+    let sum_out = out_data.iter().fold(0.0, |acc, x| acc + *x);
+    out_data = out_data
+        .par_iter()
+        .map(|a| a / sum_out)
+        .collect::<Vec<f64>>();
+
+    out_data
+}
+
+/// Generates expected values from a Poisson distribution for a single event envelope.
+pub fn fish(rate: f64, t: f64) -> f64 {
+    let mut val = -rate * t;
+    val = val.exp() * rate * t;
+    val
+}
+
+/// Goodness of fit test suite.
+pub fn gof(obs: &[f64], other: &[f64]) -> Vec<f64> {
+    let cdf = cdf_dual(obs, other);
+    let ad = ad_dual(obs, other);
+    // anderson-darling test
+    let lnx = obs.len();
+    let lny = other.len();
+    let k = lnx + lny;
+    let ada = (2.492 - 1.0) * (1.0 - (1.55 / k as f64)) + 1.0;
+    // chi-squared pearsons test
+    let ch = chi_squared(obs, other);
+    // kuiper test
+    let kp1 = cdf.iter().map(|x| x.0 - x.1).fold(0.0, f64::max);
+    let kp2 = cdf.iter().map(|x| x.1 - x.0).fold(0.0, f64::max);
+    let kp = kp1 + kp2;
+    // kolmogorov-smirnov test
+    let ks = cdf
+        .iter()
+        .map(|x| rand_distr::num_traits::abs(x.0 - x.1))
+        .fold(0.0, f64::max);
+    let ksa = 1.35810 * f64::sqrt(k as f64 / (lnx * lny) as f64); // k-s crit value at 0.05
+    vec![ad, ada, ch, kp, ks, ksa]
+}
+
 /// Produce integer-like index of f64 values from generic range.
 ///
 /// # Examples
@@ -489,3 +579,20 @@ pub fn record<T: Serialize>(rec: &mut Vec<T>, path: &str) -> Result<(), errors::
     wtr.flush()?;
     Ok(())
 }
+
+/// Timed wrapper for functions.
+pub fn timed<T: Serialize>(
+    f: fn() -> Result<Vec<T>, errors::ResError>,
+    path: &str,
+    dur: u64,
+) -> Result<(), errors::ResError> {
+    let mut rec = Vec::new();
+    let dur = std::time::Duration::new(60 * 60 * dur, 0);
+    let now = std::time::SystemTime::now();
+    while std::time::SystemTime::now() < now + dur {
+        let res = f()?;
+        rec.push(res);
+        record(&mut rec, path)?;
+    }
+    Ok(())
+}
